/* Reflow Oven Controller
 *
 * Copyright (C) 2020  Mario Hüttel <mario.huettel@gmx.net>
 *
 * This file is part of the Reflow Oven Controller Project.
 *
 * The reflow oven controller is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * The Reflow Oven Control Firmware is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the reflow oven controller project.
 * If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file main.c
 * @brief Main file for firmware
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>
/* #include <arm_math.h> */
#include <stm32/stm32f4xx.h>
#include <cmsis/core_cm4.h>
#include <setup/system_stm32f4xx.h>
#include <reflow-controller/systick.h>
#include <reflow-controller/adc-meas.h>
#include <reflow-controller/shell.h>
#include <reflow-controller/digio.h>
#include "fatfs/shimatta_sdio_driver/shimatta_sdio.h"
#include <stm-periph/stm32-gpio-macros.h>
#include <stm-periph/clock-enable-manager.h>
#include <stm-periph/uart.h>
#include <reflow-controller/shell-uart-config.h>
#include <reflow-controller/oven-driver.h>
#include <reflow-controller/safety/safety-adc.h>
#include <fatfs/ff.h>
#include <reflow-controller/reflow-menu.h>

bool global_error_state;

static void setup_nvic_priorities(void)
{
	/* No sub priorities */
	NVIC_SetPriorityGrouping(2);

	/* Setup Priorities */
	NVIC_SetPriority(ADC_IRQn, 2);
	NVIC_SetPriority(DMA2_Stream0_IRQn, 1);
	NVIC_SetPriority(DMA2_Stream7_IRQn, 3);
}

FATFS fs;
FATFS * const fs_ptr = &fs;

static inline void uart_gpio_config(void)
{
	/*
	 * In case the application is build in debug mode, use the TX/RX Pins on the debug header
	 * else the Pins on the DIGIO header are configured in the digio module
	 */

#ifdef DEBUGBUILD
	rcc_manager_enable_clock(&RCC->AHB1ENR, BITMASK_TO_BITNO(SHELL_UART_PORT_RCC_MASK));
	SHELL_UART_PORT->MODER &=  MODER_DELETE(SHELL_UART_TX_PIN) & MODER_DELETE(SHELL_UART_RX_PIN);
	SHELL_UART_PORT->MODER |= ALTFUNC(SHELL_UART_RX_PIN) | ALTFUNC(SHELL_UART_TX_PIN);
	SETAF(SHELL_UART_PORT, SHELL_UART_RX_PIN, SHELL_UART_RX_PIN_ALTFUNC);
	SETAF(SHELL_UART_PORT, SHELL_UART_TX_PIN, SHELL_UART_TX_PIN_ALTFUNC);

	/* Setup Pullup resistor at UART RX */
	SHELL_UART_PORT->PUPDR |= PULLUP(SHELL_UART_RX_PIN);
#endif
}


static char shell_uart_tx_buff[128];
static char shell_uart_rx_buff[48];
struct stm_uart shell_uart;

static shellmatta_retCode_t write_shell_callback(const char *data, uint32_t len)
{
	uart_send_array_with_dma(&shell_uart, data, len);
	return SHELLMATTA_OK;
}

static inline void setup_shell_uart(struct stm_uart *uart)
{
	uart->rx = 1;
	uart->tx = 1;
	uart->brr_val = SHELL_UART_BRR_REG_VALUE;
	uart->rcc_reg = &SHELL_UART_RCC_REG;
	uart->rcc_bit_no = BITMASK_TO_BITNO(SHELL_UART_RCC_MASK);
	uart->uart_dev = SHELL_UART_PERIPH;
	uart->dma_rx_buff = shell_uart_rx_buff;
	uart->dma_tx_buff = shell_uart_tx_buff;
	uart->rx_buff_count = sizeof(shell_uart_rx_buff);
	uart->tx_buff_count = sizeof(shell_uart_tx_buff);
	uart->base_dma_num = 2;
	uart->dma_rx_stream = SHELL_UART_RECEIVE_DMA_STREAM;
	uart->dma_tx_stream = SHELL_UART_SEND_DMA_STREAM;
	uart->dma_rx_trigger_channel = SHELL_UART_RX_DMA_TRIGGER;
	uart->dma_tx_trigger_channel = SHELL_UART_TX_DMA_TRIGGER;

	uart_init(uart);
	NVIC_EnableIRQ(DMA2_Stream7_IRQn);
}

static bool mount_sd_card_if_avail(bool mounted)
{
	FRESULT res;

	if (sdio_check_inserted() && mounted) {
		memset(fs_ptr, 0, sizeof(FATFS));
		return false;
	}

	if (!sdio_check_inserted() && !mounted) {
		res = f_mount(fs_ptr, "0:/", 1);
		if (res == FR_OK)
			return true;
		else
			return false;
	}

	return mounted;
}

static void setup_unused_pins(void)
{
	int i;

	rcc_manager_enable_clock(&RCC->AHB1ENR, BITMASK_TO_BITNO(RCC_AHB1ENR_GPIOEEN));
	GPIOE->MODER = 0UL;
	for (i = 0; i < 16; i++)
		GPIOE->PUPDR |= PULLDOWN(i);
}

static inline void setup_system(void)
{
	setup_nvic_priorities();
	systick_setup();

	adc_pt1000_setup_meas();
	oven_driver_init();
	digio_setup_default_all();
	led_setup();
	loudspeaker_setup();
	reflow_menu_init();
	safety_adc_init();

	uart_gpio_config();
	setup_shell_uart(&shell_uart);

	setup_unused_pins();
}

static void handle_shell_uart_input(shellmatta_handle_t shell_handle)
{
	int uart_receive_status;
	const char *uart_input;
	size_t uart_input_len;

	/* Handle UART input for shell */
	uart_receive_status = uart_receive_data_with_dma(&shell_uart, &uart_input, &uart_input_len);
	if (uart_receive_status >= 0)
		shell_handle_input(shell_handle, uart_input, uart_input_len);
}

static void zero_ccm_ram(void)
{
	/* These extern variables are placed in the linker script */
	extern char _sccmram;
	extern char _eccmram;
	uint32_t len;
	uint32_t i;
	uint32_t *ptr = (uint32_t *)&_sccmram;

	len = (uint32_t)&_eccmram - (uint32_t)&_sccmram;
	for (i = 0; i < len; i++)
		ptr[i] = 0UL;
}

/**
 * @brief This function sets the appropriate error flags in the oven PID controller
 * depending on the Safety ADC measurements.
 * The PID controller's error flags have to be cleared via the GUI by either starting a new RUN or explicitly
 * ack'ing these errors.
 */
static void propagate_safety_adc_error_to_oven_pid(void)
{
	enum safety_adc_check_result safety_adc_result;

	safety_adc_result = safety_adc_get_errors();

	if (safety_adc_result & SAFETY_ADC_CHECK_TEMP_LOW ||
	    safety_adc_result & SAFETY_ADC_CHECK_TEMP_HIGH)
		oven_pid_report_error(OVEN_PID_ERR_OVERTEMP);

	if (safety_adc_result & SAFETY_ADC_CHECK_VREF_LOW ||
	    safety_adc_result & SAFETY_ADC_CHECK_VREF_HIGH)
		oven_pid_report_error(OVEN_PID_ERR_VREF_TOL);

	if (safety_adc_result & SAFETY_ADC_INTERNAL_ERROR)
		oven_pid_report_error(0);
}

int main(void)
{
	bool sd_card_mounted = false;
	shellmatta_handle_t shell_handle;
	int menu_wait_request;
	uint64_t quarter_sec_timestamp = 0ULL;
	const struct oven_pid_status *pid_status;
	enum adc_pt1000_error pt1000_status;

	zero_ccm_ram();
	setup_system();

	global_error_state = false;

	shell_handle = shell_init(write_shell_callback);
	shell_print_motd(shell_handle);

	while (1) {
		sd_card_mounted = mount_sd_card_if_avail(sd_card_mounted);

		pid_status = oven_pid_get_status();

		if (systick_ticks_have_passed(quarter_sec_timestamp, 250)) {
			quarter_sec_timestamp = systick_get_global_tick();

			(void)handle_safety_adc();
			propagate_safety_adc_error_to_oven_pid();

			if (global_error_state)
				led_set(0, !led_get(0));
			else
				led_set(0, 0);
		}

		pt1000_status = adc_pt1000_check_error();
		global_error_state = pid_status->error_set || !!safety_adc_get_errors() || !!pt1000_status;

		menu_wait_request = reflow_menu_handle();

		/* Deactivate oven output in case of error! */
		if (!pid_status->active || global_error_state)
			oven_driver_set_power(0U);

		handle_shell_uart_input(shell_handle);

		if (menu_wait_request)
			__WFI();
		else
			__NOP();

	}

	return 0;
}

void sdio_wait_ms(uint32_t ms)
{
	systick_wait_ms(ms);
}

void DMA2_Stream7_IRQHandler(void)
{
	uint32_t hisr = DMA2->HISR;

	DMA2->HIFCR = hisr;

	if (hisr & DMA_HISR_TCIF7)
		uart_tx_dma_complete_int_callback(&shell_uart);
}
